"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
/* eslint-disable functional/prefer-immutable-types */
const type_utils_1 = require("@typescript-eslint/type-utils");
const utils_1 = require("@typescript-eslint/utils");
const typescript_1 = require("typescript");
const typescript_2 = require("typescript");
const common_1 = require("./common");
/**
 * An ESLint rule to ban hidden type assertions.
 */
// eslint-disable-next-line total-functions/no-unsafe-readonly-mutable-assignment
const noHiddenTypeAssertions = (0, common_1.createRule)({
    name: "no-hidden-type-assertions",
    meta: {
        type: "problem",
        docs: {
            description: "Bans hidden type assertions.",
            recommended: "error",
        },
        messages: {
            errorStringGeneric: "Do not use hidden type assertions. Specify `unknown` instead of an arbitrary type argument.",
        },
        schema: [],
    },
    create: (context) => {
        const parserServices = utils_1.ESLintUtils.getParserServices(context);
        const checker = parserServices.program.getTypeChecker();
        const explodeTypeNode = (type, depth
        // eslint-disable-next-line sonarjs/cognitive-complexity
        ) => {
            // TODO write a test that exercises this
            // eslint-disable-next-line functional/no-conditional-statements
            if (depth >= 100) {
                return [];
            }
            const next = (0, typescript_2.isTypeReferenceNode)(type)
                ? type.typeArguments ?? []
                : (0, typescript_2.isConditionalTypeNode)(type)
                    ? [type.checkType, type.trueType, type.falseType]
                    : (0, typescript_2.isIndexedAccessTypeNode)(type)
                        ? [type.objectType, type.indexType]
                        : (0, typescript_1.isFunctionTypeNode)(type)
                            ? [type.type, ...parametersToTypeNodes(type.parameters, depth)] // TODO type params?
                            : (0, typescript_1.isTypeLiteralNode)(type)
                                ? [] // TODO support type literals
                                : (0, typescript_1.isArrayTypeNode)(type)
                                    ? [type.elementType]
                                    : [];
            return [type, ...next.flatMap(explodeTypeNode)];
        };
        const parametersToTypeNodes = (parameters, depth) => {
            return parameters
                .flatMap((parameter) => parameter.type !== undefined ? [parameter.type] : [])
                .flatMap((parameter) => explodeTypeNode(parameter, depth));
        };
        return {
            // eslint-disable-next-line functional/no-return-void
            CallExpression: (node) => {
                const tsExpressionNode = parserServices.esTreeNodeToTSNodeMap.get(node);
                const callSignature = checker.getResolvedSignature(tsExpressionNode);
                // eslint-disable-next-line functional/no-conditional-statements
                if (callSignature === undefined ||
                    callSignature.declaration === undefined) {
                    return;
                }
                // eslint-disable-next-line functional/no-conditional-statements
                if (callSignature.declaration.kind === typescript_2.SyntaxKind.JSDocSignature) {
                    // TODO support JSDoc
                    return;
                }
                const typeParameters = callSignature.declaration.typeParameters;
                const parameters = parametersToTypeNodes(callSignature.declaration.parameters, 0);
                const returnType = callSignature.declaration.type;
                // eslint-disable-next-line functional/no-conditional-statements
                if (returnType === undefined ||
                    typeParameters === undefined ||
                    typeParameters.length === 0) {
                    return;
                }
                const returnTypes = explodeTypeNode(returnType, 0);
                // Of all the type parameters, these are the ones that are used in the
                // return type (alongside their positional index, which we use next to
                // look up corresponding type arguments)
                const typeParamsUsedInReturnType = typeParameters
                    .map((typeParameter, index) => ({ typeParameter, index }))
                    .filter((typeParameter) => returnTypes.some((returnType) => (0, typescript_2.isTypeReferenceNode)(returnType) &&
                    returnType.typeName.kind === typescript_2.SyntaxKind.Identifier &&
                    returnType.typeName.text ===
                        typeParameter.typeParameter.name.text));
                // Of all the type parameters that appear in the return type,
                // are they all set to `unknown` type arguments in this specific call?
                // If so, this is safe even if the function being called is a hidden type assertion.
                const allCorrespondingTypeArgumentsAreUnknownType = typeParamsUsedInReturnType.every(({ index }) => {
                    const typeArgument = (tsExpressionNode.typeArguments ?? [])[index];
                    return (typeArgument !== undefined &&
                        ((0, type_utils_1.isTypeUnknownType)(checker.getTypeAtLocation(typeArgument)) ||
                            (0, type_utils_1.isTypeNeverType)(checker.getTypeAtLocation(typeArgument))));
                });
                // eslint-disable-next-line functional/no-conditional-statements
                if (allCorrespondingTypeArgumentsAreUnknownType) {
                    return;
                }
                // Confirms that all the type parameters that appear in the return type ALSO
                // appear in at least one (value) parameter. If they do, this probably isn't
                // a hidden type assertion.
                const allTypeParamsUsedInReturnTypeAlsoAppearInValueParams = typeParamsUsedInReturnType
                    .map(({ typeParameter }) => typeParameter)
                    .every((typeParameter) => {
                    return parameters.some((parameter) => {
                        return ((0, typescript_2.isTypeReferenceNode)(parameter) &&
                            parameter.typeName.kind === typescript_2.SyntaxKind.Identifier &&
                            typeParameter.name.text === parameter.typeName.text);
                    });
                });
                // eslint-disable-next-line functional/no-conditional-statements
                if (!allTypeParamsUsedInReturnTypeAlsoAppearInValueParams) {
                    // eslint-disable-next-line functional/no-expression-statements
                    context.report({
                        node: node,
                        messageId: "errorStringGeneric",
                    });
                }
            },
        };
    },
    defaultOptions: [],
});
exports.default = noHiddenTypeAssertions;
